#
#   Copyright 2023 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import json
import os
from typing import Optional

from hopsworks_common import client, decorators, flink_cluster, job, usage, util
from hopsworks_common.client.exceptions import RestAPIError
from hopsworks_common.core import job_api


class FlinkClusterApi:
    def __init__(self):
        self._job_api = job_api.JobApi()

    @usage.method_logger
    def get_configuration(self):
        """Get configuration for the Flink cluster.

        # Returns
            `dict`: Default configuration for the Flink cluster,
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        return self._job_api.get_configuration("FLINK")

    @usage.method_logger
    def setup_cluster(self, name: str, config=None):
        """Create a new flink job representing a flink cluster, or update an existing one.

        ```python

        import hopsworks

        project = hopsworks.login()

        flink_cluster_api = project.get_flink_cluster_api()

        flink_config = flink_cluster_api.get_configuration()

        flink_config['appName'] = "myFlinkCluster"

        flink_cluster = flink_cluster_api.setup_cluster(name="myFlinkCluster", config=flink_config)
        ```
        # Arguments
            name: Name of the cluster.
            config: Configuration of the cluster.
        # Returns
            `FlinkCluster`: The FlinkCluster object representing the cluster
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        if self._job_api.exists(name):
            # If the job already exists, retrieve it
            _flink_cluster = self.get_cluster(name)
            if _flink_cluster._job.job_type != "FLINK":
                raise RestAPIError(
                    "This is not a Flink cluster. Please use different name to create new Flink cluster"
                )
            return _flink_cluster
        else:
            # If the job doesn't exists, create a new job
            if config is None:
                config = self.get_configuration()
                config["appName"] = name
            return self._create_cluster(name, config)

    def _create_cluster(self, name: str, config: dict):
        _client = client.get_instance()

        config = util.validate_job_conf(config, _client._project_name)

        path_params = ["project", _client._project_id, "jobs", name]

        headers = {"content-type": "application/json"}
        flink_job = job.Job.from_response_json(
            _client._send_request(
                "PUT", path_params, headers=headers, data=json.dumps(config)
            )
        )
        flink_cluster_obj = flink_cluster.FlinkCluster(flink_job)
        print(
            f"FlinkCluster created successfully, explore it at {flink_cluster_obj.get_url()}"
        )
        return flink_cluster_obj

    @usage.method_logger
    @decorators.catch_not_found(
        "hopsworks_common.flink_cluster.FlinkCluster", fallback_return=None
    )
    def get_cluster(self, name: str) -> Optional[flink_cluster.FlinkCluster]:
        """Get the job corresponding to the flink cluster.
        ```python

        import hopsworks

        project = hopsworks.login()

        flink_cluster_api = project.get_flink_cluster_api()

        flink_cluster = flink_cluster_api.get_cluster(name="myFlinkCluster")
        ```

        # Arguments
            name: Name of the cluster.
        # Returns
            `FlinkCluster`: The FlinkCluster object representing the cluster or `None` if it does not exist.
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "jobs",
            name,
        ]
        query_params = {"expand": ["creator"]}
        flink_job = job.Job.from_response_json(
            _client._send_request("GET", path_params, query_params=query_params)
        )
        return flink_cluster.FlinkCluster(flink_job)

    def _get_job(self, execution, job_id):
        """Get specific job from the specific execution of the flink cluster.
        ```python

        # log in to hopsworks
        import hopsworks
        project = hopsworks.login()

        # fetch flink cluster handle
        flink_cluster_api = project.get_flink_cluster_api()
        flink_cluster = flink_cluster_api.get_cluster(name="myFlinkCluster")

        # get jobs from this execution
        job_id = '113a2af5b724a9b92085dc2d9245e1d6'
        flink_cluster_api.get_job(execution, job_id)
        ```

        # Arguments
            execution: Execution object.
            job_id: id of the job within this execution
        # Returns
            `Dict`: Dict with flink job id and and status of the job.
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """

        _client = client.get_instance()
        path_params = ["hopsworks-api", "flinkmaster", execution.app_id, "jobs", job_id]
        headers = {"content-type": "application/json"}
        return _client._send_request(
            "GET", path_params, headers=headers, with_base_path_params=False
        )

    def _get_jobs(self, execution):
        """Get jobs from the specific execution of the flink cluster.
        ```python

        # log in to hopsworks
        import hopsworks
        project = hopsworks.login()

        # fetch flink cluster handle
        flink_cluster_api = project.get_flink_cluster_api()
        flink_cluster = flink_cluster_api.get_cluster(name="myFlinkCluster")

        # get jobs from this execution
        flink_cluster_api.get_jobs(execution)
        ```

        # Arguments
            execution: Execution object.
        # Returns
            `List[Dict]`: The array of dicts with flink job id and and status of the job.
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """

        _client = client.get_instance()
        path_params = ["hopsworks-api", "flinkmaster", execution.app_id, "jobs"]
        headers = {"content-type": "application/json"}
        return _client._send_request(
            "GET", path_params, headers=headers, with_base_path_params=False
        )["jobs"]

    def _stop_execution(self, execution):
        """Stop specific execution of the flink cluster.
        ```python

        # log in to hopsworks
        import hopsworks
        project = hopsworks.login()

        # fetch flink cluster handle
        flink_cluster_api = project.get_flink_cluster_api()
        flink_cluster = flink_cluster_api.get_cluster(name="myFlinkCluster")

        # stop this execution
        flink_cluster_api.stop_execution(execution)
        ```

        # Arguments
            execution: Execution object.
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        _client = client.get_instance()
        path_params = ["hopsworks-api", "flinkmaster", execution.app_id, "cluster"]
        headers = {"content-type": "application/json"}
        _client._send_request(
            "DELETE", path_params, headers=headers, with_base_path_params=False
        )

    def _stop_job(self, execution, job_id):
        """Stop specific job of the specific execution of the flink cluster.
        ```python

        # log in to hopsworks
        import hopsworks
        project = hopsworks.login()

        # fetch flink cluster handle
        flink_cluster_api = project.get_flink_cluster_api()
        flink_cluster = flink_cluster_api.get_cluster(name="myFlinkCluster")

        # stop the job
        job_id = '113a2af5b724a9b92085dc2d9245e1d6'
        flink_cluster_api.stop_job(execution, job_id)
        ```

        # Arguments
            execution: Execution object.
            job_id: id of the job within this execution
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        _client = client.get_instance()
        path_params = [
            "hopsworks-api",
            "flinkmaster",
            execution.app_id,
            "jobs",
            job_id,
        ]
        headers = {"content-type": "application/json"}
        return _client._send_request(
            "PATCH", path_params, headers=headers, with_base_path_params=False
        )

    def _get_jars(self, execution):
        """Get already uploaded jars from the specific execution of the flink cluster.
        ```python
        # log in to hopsworks
        import hopsworks
        project = hopsworks.login()

        # fetch flink cluster handle
        flink_cluster_api = project.get_flink_cluster_api()
        flink_cluster = flink_cluster_api.get_cluster(name="myFlinkCluster")

        # get jar files from this execution
        flink_cluster_api.get_jars(execution)
        ```

        # Arguments
            execution: Execution object.
        # Returns
            `List[Dict]`: The array of dicts with jar metadata.
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """

        _client = client.get_instance()
        path_params = ["hopsworks-api", "flinkmaster", execution.app_id, "jars"]
        headers = {"content-type": "application/json"}
        response = _client._send_request(
            "GET", path_params, headers=headers, with_base_path_params=False
        )
        return response["files"]

    def _upload_jar(self, execution, jar_file):
        """Uploaded jar file to the specific execution of the flink cluster.
        ```python
        # log in to hopsworks
        import hopsworks
        project = hopsworks.login()

        # fetch flink cluster handle
        flink_cluster_api = project.get_flink_cluster_api()
        flink_cluster = flink_cluster_api.get_cluster(name="myFlinkCluster")

        # upload jar file jobs from this execution
        jar_file_path = "./flink-example.jar"
        flink_cluster_api.upload_jar(execution, jar_file_path)
        ```

        # Arguments
            execution: Execution object.
            jar_file: path to the jar file.
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """

        _client = client.get_instance()
        path_params = [
            "hopsworks-api",
            "flinkmaster",
            execution.app_id,
            "jars",
            "upload",
        ]
        files = {
            "jarfile": (
                os.path.basename(jar_file),
                open(jar_file, "rb"),
                "application/x-java-archive",
            )
        }
        _client._send_request(
            "POST", path_params, files=files, with_base_path_params=False
        )
        print("Flink Jar uploaded.")

    def _submit_job(self, execution, jar_id, main_class, job_arguments):
        """Submit job using the specific jar file, already uploaded to this execution of the flink cluster.
        ```python
        # log in to hopsworks
        import hopsworks
        project = hopsworks.login()

        # fetch flink cluster handle
        flink_cluster_api = project.get_flink_cluster_api()
        flink_cluster = flink_cluster_api.get_cluster(name="myFlinkCluster")

        # upload jar file jobs from this execution
        main_class = "com.example.Main"
        job_arguments = "-arg1 arg1 -arg2 arg2"

        #get jar file metadata (and select the 1st one for demo purposes)
        jar_metadata = flink_cluster_api.get_jars(execution)[0]
        jar_id = jar_metadata["id"]
        flink_cluster_api.submit_job(execution, jar_id, main_class, job_arguments=job_arguments)
        ```

        # Arguments
            execution: Execution object.
            jar_id: id if the jar file
            main_class: path to the main class of the the jar file
            job_arguments: Job arguments (if any), defaults to none.
        # Returns
            `str`:  job id.
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """

        _client = client.get_instance()
        # Submit execution
        if job_arguments:
            path_params = [
                "hopsworks-api",
                "flinkmaster",
                execution.app_id,
                "jars",
                jar_id,
                f"run?entry-class={main_class}&program-args={job_arguments}",
            ]
        else:
            path_params = [
                "hopsworks-api",
                "flinkmaster",
                execution.app_id,
                "jars",
                jar_id,
                f"run?entry-class{main_class}",
            ]

        headers = {"content-type": "application/json"}
        response = _client._send_request(
            "POST", path_params, headers=headers, with_base_path_params=False
        )

        job_id = response["jobid"]
        print("Submitted Job Id: {}".format(job_id))

        return job_id

    def _job_state(self, execution, job_id):
        """Gets state of the job from the specific execution of the flink cluster.
        ```python

        # log in to hopsworks
        import hopsworks
        project = hopsworks.login()

        # fetch flink cluster handle
        flink_cluster_api = project.get_flink_cluster_api()
        flink_cluster = flink_cluster_api.get_cluster(name="myFlinkCluster")

        # get jobs from this execution
        job_id = '113a2af5b724a9b92085dc2d9245e1d6'
        flink_cluster_api.job_status(execution, job_id)
        ```

        # Arguments
            execution: Execution object.
        # Returns
            `str`: status of the job. Possible states:  "INITIALIZING", "CREATED", "RUNNING", "FAILING", "FAILED",
            "CANCELLING", "CANCELED",  "FINISHED", "RESTARTING", "SUSPENDED", "RECONCILING".
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: If the backend encounters an error when handling the request
        """
        _client = client.get_instance()
        path_params = [
            "hopsworks-api",
            "flinkmaster",
            execution.app_id,
            "jobs",
            job_id,
        ]
        response = _client._send_request(
            "GET", path_params, with_base_path_params=False
        )

        return response["state"]



#
#   Copyright 2021 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import json
import logging
import os
import re
import warnings
from typing import Any, Dict, Optional, Union

import humps
from hopsworks_common import client, usage, util
from hopsworks_common.constants import ARTIFACT_VERSION, MODEL_REGISTRY
from hopsworks_common.constants import INFERENCE_ENDPOINTS as IE
from hsml import deployment, tag
from hsml.core import explicit_provenance
from hsml.engine import model_engine
from hsml.inference_batcher import InferenceBatcher
from hsml.inference_logger import InferenceLogger
from hsml.model_schema import ModelSchema
from hsml.predictor import Predictor
from hsml.resources import PredictorResources
from hsml.schema import Schema
from hsml.transformer import Transformer


_logger = logging.getLogger(__name__)


class Model:
    NOT_FOUND_ERROR_CODE = 360000
    """Metadata object representing a model in the Model Registry."""

    def __init__(
        self,
        id,
        name,
        version=None,
        created=None,
        creator=None,
        environment=None,
        description=None,
        project_name=None,
        metrics=None,
        program=None,
        user_full_name=None,
        model_schema=None,
        input_example=None,
        framework=None,
        model_registry_id=None,
        # unused, but needed since they come in the backend response
        tags=None,
        href=None,
        feature_view=None,
        training_dataset_version=None,
        **kwargs,
    ):
        self._id = id
        self._name = name
        self._version = version

        if description is None:
            self._description = "A collection of models for " + name
        else:
            self._description = description

        self._created = created
        self._creator = creator
        self._environment = environment
        self._project_name = project_name
        self._training_metrics = metrics
        self._program = program
        self._user_full_name = user_full_name
        self._input_example = input_example
        self._framework = framework
        self._model_schema = model_schema

        # This is needed for update_from_response_json function to not overwrite name of the shared registry this model originates from
        if not hasattr(self, "_shared_registry_project_name"):
            self._shared_registry_project_name = None

        self._model_registry_id = model_registry_id

        self._model_engine = model_engine.ModelEngine()
        self._feature_view = feature_view
        self._training_dataset_version = training_dataset_version

    @usage.method_logger
    def save(
        self,
        model_path,
        await_registration=480,
        keep_original_files=False,
        upload_configuration: Optional[Dict[str, Any]] = None,
    ):
        """Persist this model including model files and metadata to the model registry.

        # Arguments
            model_path: Local or remote (Hopsworks file system) path to the folder where the model files are located, or path to a specific model file.
            await_registration: Awaiting time for the model to be registered in Hopsworks.
            keep_original_files: If the model files are located in hopsfs, whether to move or copy those files into the Models dataset. Default is False (i.e., model files will be moved)
            upload_configuration: When saving a model from outside Hopsworks, the model is uploaded to the model registry using the REST APIs. Each model artifact is divided into
                chunks and each chunk uploaded independently. This parameter can be used to control the upload chunk size, the parallelism and the number of retries.
                `upload_configuration` can contain the following keys:
                * key `chunk_size`: size of each chunk in megabytes. Default 10.
                * key `simultaneous_uploads`: number of chunks to upload in parallel. Default 3.
                * key `max_chunk_retries`: number of times to retry the upload of a chunk in case of failure. Default 1.

        # Returns
            `Model`: The model metadata object.
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: In case the backend encounters an issue
        """
        if self._training_dataset_version is None and self._feature_view is not None:
            if self._feature_view.get_last_accessed_training_dataset() is not None:
                self._training_dataset_version = (
                    self._feature_view.get_last_accessed_training_dataset()
                )
            else:
                warnings.warn(
                    "Provenance cached data - feature view provided, but training dataset version is missing",
                    util.ProvenanceWarning,
                    stacklevel=1,
                )
        if self._model_schema is None:
            if (
                self._feature_view is not None
                and self._training_dataset_version is not None
            ):
                all_features = self._feature_view.get_training_dataset_schema(
                    self._training_dataset_version
                )
                features, labels = [], []
                for feature in all_features:
                    (labels if feature.label else features).append(feature.to_dict())
                self._model_schema = ModelSchema(
                    input_schema=Schema(features) if features else None,
                    output_schema=Schema(labels) if labels else None,
                )
            else:
                warnings.warn(
                    "Model schema cannot not be inferred without both the feature view and the training dataset version.",
                    util.ProvenanceWarning,
                    stacklevel=1,
                )

        return self._model_engine.save(
            model_instance=self,
            model_path=model_path,
            await_registration=await_registration,
            keep_original_files=keep_original_files,
            upload_configuration=upload_configuration,
        )

    @usage.method_logger
    def download(self, local_path=None) -> str:
        """Download the model files.

        # Arguments
            local_path: path where to download the model files in the local filesystem
        # Returns
            `str`: Absolute path to local folder containing the model files.
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: In case the backend encounters an issue
        """
        return self._model_engine.download(model_instance=self, local_path=local_path)

    @usage.method_logger
    def delete(self):
        """Delete the model

        !!! danger "Potentially dangerous operation"
            This operation drops all metadata associated with **this version** of the
            model **and** deletes the model files.

        # Raises
            `hopsworks.client.exceptions.RestAPIError`: In case the backend encounters an issue
        """
        self._model_engine.delete(model_instance=self)

    @usage.method_logger
    def deploy(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
        artifact_version: Optional[str] = ARTIFACT_VERSION.CREATE,
        serving_tool: Optional[str] = None,
        script_file: Optional[str] = None,
        config_file: Optional[str] = None,
        resources: Optional[Union[PredictorResources, dict]] = None,
        inference_logger: Optional[Union[InferenceLogger, dict]] = None,
        inference_batcher: Optional[Union[InferenceBatcher, dict]] = None,
        transformer: Optional[Union[Transformer, dict]] = None,
        api_protocol: Optional[str] = IE.API_PROTOCOL_REST,
        environment: Optional[str] = None,
    ) -> deployment.Deployment:
        """Deploy the model.

        !!! example
            ```python

            import hopsworks

            project = hopsworks.login()

            # get Hopsworks Model Registry handle
            mr = project.get_model_registry()

            # retrieve the trained model you want to deploy
            my_model = mr.get_model("my_model", version=1)

            my_deployment = my_model.deploy()
            ```
        # Arguments
            name: Name of the deployment.
            description: Description of the deployment.
            artifact_version: Version number of the model artifact to deploy, `CREATE` to create a new model artifact
            or `MODEL-ONLY` to reuse the shared artifact containing only the model files.
            serving_tool: Serving tool used to deploy the model server.
            script_file: Path to a custom predictor script implementing the Predict class.
            config_file: Model server configuration file to be passed to the model deployment.
                It can be accessed via `CONFIG_FILE_PATH` environment variable from a predictor or transformer script.
                For LLM deployments without a predictor script, this file is used to configure the vLLM engine.
            resources: Resources to be allocated for the predictor.
            inference_logger: Inference logger configuration.
            inference_batcher: Inference batcher configuration.
            transformer: Transformer to be deployed together with the predictor.
            api_protocol: API protocol to be enabled in the deployment (i.e., 'REST' or 'GRPC'). Defaults to 'REST'.
            environment: The inference environment to use.

        # Returns
            `Deployment`: The deployment metadata object of a new or existing deployment.

        # Raises
            `hopsworks.client.exceptions.RestAPIError`: In case the backend encounters an issue
        """

        if name is None:
            name = self._get_default_serving_name()

        predictor = Predictor.for_model(
            self,
            name=name,
            description=description,
            artifact_version=artifact_version,
            serving_tool=serving_tool,
            script_file=script_file,
            config_file=config_file,
            resources=resources,
            inference_logger=inference_logger,
            inference_batcher=inference_batcher,
            transformer=transformer,
            api_protocol=api_protocol,
            environment=environment,
        )

        return predictor.deploy()

    @usage.method_logger
    def set_tag(self, name: str, value: Union[str, dict]):
        """Attach a tag to a model.

        A tag consists of a <name,value> pair. Tag names are unique identifiers across the whole cluster.
        The value of a tag can be any valid json - primitives, arrays or json objects.

        # Arguments
            name: Name of the tag to be added.
            value: Value of the tag to be added.
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: in case the backend fails to add the tag.
        """

        self._model_engine.set_tag(model_instance=self, name=name, value=value)

    @usage.method_logger
    def delete_tag(self, name: str):
        """Delete a tag attached to a model.

        # Arguments
            name: Name of the tag to be removed.
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: in case the backend fails to delete the tag.
        """
        self._model_engine.delete_tag(model_instance=self, name=name)

    def get_tag(self, name: str) -> Optional[str]:
        """Get the tags of a model.

        # Arguments
            name: Name of the tag to get.
        # Returns
            tag value or `None` if it does not exist.
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: in case the backend fails to retrieve the tag.
        """
        return self._model_engine.get_tag(model_instance=self, name=name)

    def get_tags(self) -> Dict[str, tag.Tag]:
        """Retrieves all tags attached to a model.

        # Returns
            `Dict[str, obj]` of tags.
        # Raises
            `hopsworks.client.exceptions.RestAPIError` in case the backend fails to retrieve the tags.
        """
        return self._model_engine.get_tags(model_instance=self)

    def get_url(self):
        path = (
            "/p/"
            + str(client.get_instance()._project_id)
            + "/models/"
            + str(self.name)
            + "/"
            + str(self.version)
        )
        return util.get_hostname_replaced_url(sub_path=path)

    def get_feature_view(self, init: bool = True, online: bool = False):
        """Get the parent feature view of this model, based on explicit provenance.
         Only accessible, usable feature view objects are returned. Otherwise an Exception is raised.
         For more details, call the base method - get_feature_view_provenance

         # Arguments
            init: By default this is set to True. If you require a more complex initialization of the feature view for online or batch scenarios, you should set `init` to False to retrieve a non initialized feature view and then call `init_batch_scoring()` or `init_serving()` with the required parameters.
            online: By default this is set to False and the initialization for batch scoring is considered the default scenario. If you set `online` to True, the online scenario is enabled and the `init_serving()` method is called. When inside a deployment, the only available scenario is the online one, thus the parameter is ignored and init_serving is always called (if `init` is set to True). If you want to override this behaviour, you should set `init` to False and proceed with a custom initialization.
        # Returns
            `FeatureView`: Feature View Object or `None` if it does not exist.
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: in case the backend fails to retrieve the feature view.
        """
        fv_prov = self.get_feature_view_provenance()
        fv = explicit_provenance.Links.get_one_accessible_parent(fv_prov)
        if fv is None:
            return None
        if init:
            td_prov = self.get_training_dataset_provenance()
            td = explicit_provenance.Links.get_one_accessible_parent(td_prov)
            is_deployment = "DEPLOYMENT_NAME" in os.environ
            if online or is_deployment:
                _logger.info(
                    "Initializing for batch and online retrieval of feature vectors"
                    + (" - within a deployment" if is_deployment else "")
                )
                fv.init_serving(training_dataset_version=td.version)
            elif online is False:
                _logger.info("Initializing for batch retrieval of feature vectors")
                fv.init_batch_scoring(training_dataset_version=td.version)
        return fv

    def get_feature_view_provenance(self) -> explicit_provenance.Links:
        """Get the parent feature view of this model, based on explicit provenance.
        This feature view can be accessible, deleted or inaccessible.
        For deleted and inaccessible feature views, only a minimal information is
        returned.

        # Returns
            `Links`: Object containing the section of provenance graph requested or `None` if it does not exist
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: in case the backend fails to retrieve the feature view provenance
        """
        return self._model_engine.get_feature_view_provenance(model_instance=self)

    def get_training_dataset_provenance(self) -> explicit_provenance.Links:
        """Get the parent training dataset of this model, based on explicit provenance.
        This training dataset can be accessible, deleted or inaccessible.
        For deleted and inaccessible training datasets, only a minimal information is
        returned.

        # Returns
            `Links`: Object containing the section of provenance graph requested or `None` if it does not exist
        # Raises
            `hopsworks.client.exceptions.RestAPIError`: in case the backend fails to retrieve the training dataset provenance
        """
        return self._model_engine.get_training_dataset_provenance(model_instance=self)

    def _get_default_serving_name(self):
        return re.sub(r"[^a-zA-Z0-9]", "", self._name)

    @classmethod
    def from_response_json(cls, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        if "count" in json_decamelized:
            if json_decamelized["count"] == 0:
                return []
            return [util.set_model_class(model) for model in json_decamelized["items"]]
        else:
            return util.set_model_class(json_decamelized)

    def update_from_response_json(self, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        if "type" in json_decamelized:  # backwards compatibility
            _ = json_decamelized.pop("type")
        self.__init__(**json_decamelized)
        return self

    def json(self):
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self):
        return {
            "id": self._name + "_" + str(self._version),
            "projectName": self._project_name,
            "name": self._name,
            "modelSchema": self._model_schema,
            "version": self._version,
            "description": self._description,
            "inputExample": self._input_example,
            "framework": self._framework,
            "metrics": self._training_metrics,
            "environment": self._environment,
            "program": self._program,
            "featureView": util.feature_view_to_json(self._feature_view),
            "trainingDatasetVersion": self._training_dataset_version,
        }

    @property
    def id(self):
        """Id of the model."""
        return self._id

    @id.setter
    def id(self, id):
        self._id = id

    @property
    def name(self):
        """Name of the model."""
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    @property
    def version(self):
        """Version of the model."""
        return self._version

    @version.setter
    def version(self, version):
        self._version = version

    @property
    def description(self):
        """Description of the model."""
        return self._description

    @description.setter
    def description(self, description):
        self._description = description

    @property
    def created(self):
        """Creation date of the model."""
        return self._created

    @created.setter
    def created(self, created):
        self._created = created

    @property
    def creator(self):
        """Creator of the model."""
        return self._creator

    @creator.setter
    def creator(self, creator):
        self._creator = creator

    @property
    def environment(self):
        """Input example of the model."""
        if self._environment is not None:
            return self._model_engine.read_file(
                model_instance=self, resource="environment.yml"
            )
        return self._environment

    @environment.setter
    def environment(self, environment):
        self._environment = environment

    @property
    def training_metrics(self):
        """Training metrics of the model."""
        return self._training_metrics

    @training_metrics.setter
    def training_metrics(self, training_metrics):
        self._training_metrics = training_metrics

    @property
    def program(self):
        """Executable used to export the model."""
        if self._program is not None:
            return self._model_engine.read_file(
                model_instance=self, resource=self._program
            )

    @program.setter
    def program(self, program):
        self._program = program

    @property
    def user(self):
        """user of the model."""
        return self._user_full_name

    @user.setter
    def user(self, user_full_name):
        self._user_full_name = user_full_name

    @property
    def input_example(self):
        """input_example of the model."""
        return self._model_engine.read_json(
            model_instance=self, resource="input_example.json"
        )

    @input_example.setter
    def input_example(self, input_example):
        self._input_example = input_example

    @property
    def framework(self):
        """framework of the model."""
        return self._framework

    @framework.setter
    def framework(self, framework):
        self._framework = framework

    @property
    def model_schema(self):
        """model schema of the model."""
        return self._model_engine.read_json(
            model_instance=self, resource="model_schema.json"
        )

    @model_schema.setter
    def model_schema(self, model_schema):
        self._model_schema = model_schema

    @property
    def project_name(self):
        """project_name of the model."""
        return self._project_name

    @project_name.setter
    def project_name(self, project_name):
        self._project_name = project_name

    @property
    def model_registry_id(self):
        """model_registry_id of the model."""
        return self._model_registry_id

    @model_registry_id.setter
    def model_registry_id(self, model_registry_id):
        self._model_registry_id = model_registry_id

    @property
    def model_path(self):
        """path of the model with version folder omitted. Resolves to /Projects/{project_name}/Models/{name}"""
        return "/Projects/{}/Models/{}".format(self.project_name, self.name)

    @property
    def version_path(self):
        """path of the model including version folder. Resolves to /Projects/{project_name}/Models/{name}/{version}"""
        return "{}/{}".format(self.model_path, str(self.version))

    @property
    def model_files_path(self):
        """path of the model files including version and files folder. Resolves to /Projects/{project_name}/Models/{name}/{version}/Files"""
        return "{}/{}".format(
            self.version_path,
            MODEL_REGISTRY.MODEL_FILES_DIR_NAME,
        )

    @property
    def shared_registry_project_name(self):
        """shared_registry_project_name of the model."""
        return self._shared_registry_project_name

    @shared_registry_project_name.setter
    def shared_registry_project_name(self, shared_registry_project_name):
        self._shared_registry_project_name = shared_registry_project_name

    @property
    def training_dataset_version(self) -> int:
        return self._training_dataset_version

    def __repr__(self):
        return f"Model(name: {self._name!r}, version: {self._version!r})"



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import json
from datetime import datetime
from typing import Any, Dict, List, Optional, Union

import hsfs
import humps
from hsfs import util
from hsfs.constructor import filter
from hsfs.decorators import typechecked


@typechecked
class Feature:
    """Metadata object representing a feature in a feature group in the Feature Store.

    See Training Dataset Feature for the
    feature representation of training dataset schemas.
    """

    COMPLEX_TYPES = ["MAP", "ARRAY", "STRUCT", "UNIONTYPE"]

    def __init__(
        self,
        name: str,
        type: Optional[str] = None,
        description: Optional[str] = None,
        primary: bool = False,
        foreign: bool = False,
        partition: bool = False,
        hudi_precombine_key: bool = False,
        online_type: Optional[str] = None,
        default_value: Optional[str] = None,
        feature_group_id: Optional[int] = None,
        feature_group: Optional[
            Union[
                "hsfs.feature_group.FeatureGroup",
                "hsfs.feature_group.ExternalFeatureGroup",
                "hsfs.feature_group.SpineGroup",
            ]
        ] = None,
        on_demand: bool = False,
        use_fully_qualified_name=False,
        **kwargs,
    ) -> None:
        self._name = util.autofix_feature_name(name)
        self._type = type
        self._description = description
        self._primary = primary
        self._foreign = foreign
        self._partition = partition
        self._hudi_precombine_key = hudi_precombine_key
        self._online_type = online_type
        self._default_value = default_value
        self._use_fully_qualified_name = use_fully_qualified_name
        if feature_group is not None:
            self._feature_group_id = feature_group.id
        else:
            self._feature_group_id = feature_group_id
        self._on_demand = on_demand

    def to_dict(self) -> Dict[str, Any]:
        """Get structured info about specific Feature in python dictionary format.

        !!! example
            ```python
            # connect to the Feature Store
            fs = ...

            # get the Feature Group instance
            fg = fs.get_or_create_feature_group(...)

            selected_feature = fg.get_feature("min_temp")
            selected_feature.to_dict()
            ```
        """
        return {
            "name": self._name,
            "type": self._type,
            "description": self._description,
            "partition": self._partition,
            "hudiPrecombineKey": self._hudi_precombine_key,
            "primary": self._primary,
            "foreign": self._foreign,
            "onlineType": self._online_type,
            "defaultValue": self._default_value,
            "featureGroupId": self._feature_group_id,
            "onDemand": self.on_demand,
            "useFullyQualifiedName": self._use_fully_qualified_name,
        }

    def json(self) -> str:
        return json.dumps(self, cls=util.Encoder)

    @classmethod
    def from_response_json(cls, json_dict: Dict[str, Any]) -> "Feature":
        if json_dict is None:
            return None

        json_decamelized = humps.decamelize(json_dict)
        return cls(**json_decamelized)

    def is_complex(self) -> bool:
        """Returns true if the feature has a complex type.

        !!! example
            ```python
            # connect to the Feature Store
            fs = ...

            # get the Feature Group instance
            fg = fs.get_or_create_feature_group(...)

            selected_feature = fg.get_feature("min_temp")
            selected_feature.is_complex()
            ```
        """
        return any(map(self._type.upper().startswith, self.COMPLEX_TYPES))

    @property
    def name(self) -> str:
        """Name of the feature."""
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        self._name = name

    @property
    def description(self) -> Optional[str]:
        """Description of the feature."""
        return self._description

    @description.setter
    def description(self, description: Optional[str]) -> None:
        self._description = description

    @property
    def type(self) -> Optional[str]:
        """Data type of the feature in the offline feature store.

        !!! danger "Not a Python type"
            This type property is not to be confused with Python types.
            The type property represents the actual data type of the feature in
            the feature store.
        """
        return self._type

    @type.setter
    def type(self, type: Optional[str]) -> None:
        self._type = type

    @property
    def online_type(self) -> Optional[str]:
        """Data type of the feature in the online feature store."""
        return self._online_type

    @online_type.setter
    def online_type(self, online_type: Optional[str]) -> None:
        self._online_type = online_type

    @property
    def primary(self) -> bool:
        """Whether the feature is part of the primary key of the feature group."""
        return self._primary

    @primary.setter
    def primary(self, primary: bool) -> None:
        self._primary = primary

    @property
    def foreign(self) -> bool:
        """Whether the feature is part of the foreign key of the feature group."""
        return self._foreign

    @foreign.setter
    def foreign(self, foreign: bool) -> None:
        self._foreign = foreign

    @property
    def partition(self) -> bool:
        """Whether the feature is part of the partition key of the feature group."""
        return self._partition

    @partition.setter
    def partition(self, partition: bool) -> None:
        self._partition = partition

    @property
    def hudi_precombine_key(self) -> bool:
        """Whether the feature is part of the hudi precombine key of the feature group."""
        return self._hudi_precombine_key

    @hudi_precombine_key.setter
    def hudi_precombine_key(self, hudi_precombine_key: bool) -> None:
        self._hudi_precombine_key = hudi_precombine_key

    @property
    def default_value(self) -> Optional[str]:
        """Default value of the feature as string, if the feature was appended to the
        feature group."""
        return self._default_value

    @default_value.setter
    def default_value(self, default_value: Optional[str]) -> None:
        self._default_value = default_value

    @property
    def feature_group_id(self) -> Optional[int]:
        return self._feature_group_id

    @property
    def on_demand(self) -> bool:
        """Whether the feature is a on-demand feature computed using on-demand transformation functions"""
        return self._on_demand

    @on_demand.setter
    def on_demand(self, on_demand) -> None:
        self._on_demand = on_demand

    def _get_filter_value(self, value: Any) -> Any:
        if self.type == "timestamp":
            return datetime.fromtimestamp(
                util.convert_event_time_to_timestamp(value) / 1000
            ).strftime("%Y-%m-%d %H:%M:%S")
        else:
            return value

    def __lt__(self, other: Any) -> "filter.Filter":
        return filter.Filter(self, filter.Filter.LT, self._get_filter_value(other))

    def __le__(self, other: Any) -> "filter.Filter":
        return filter.Filter(self, filter.Filter.LE, self._get_filter_value(other))

    def __eq__(self, other: Any) -> "filter.Filter":
        return filter.Filter(self, filter.Filter.EQ, self._get_filter_value(other))

    def __ne__(self, other: Any) -> "filter.Filter":
        return filter.Filter(self, filter.Filter.NE, self._get_filter_value(other))

    def __ge__(self, other: Any) -> "filter.Filter":
        return filter.Filter(self, filter.Filter.GE, self._get_filter_value(other))

    def __gt__(self, other: Any) -> "filter.Filter":
        return filter.Filter(self, filter.Filter.GT, self._get_filter_value(other))

    def contains(self, other: Union[str, List[Any]]) -> "filter.Filter":
        """
        !!! warning "Deprecated"
            `contains` method is deprecated. Use `isin` instead.
        """
        return self.isin(other)

    def isin(self, other: Union[str, List[Any]]) -> "filter.Filter":
        return filter.Filter(self, filter.Filter.IN, json.dumps(other))

    def like(self, other: Any) -> "filter.Filter":
        return filter.Filter(self, filter.Filter.LK, other)

    def __str__(self) -> str:
        return self.json()

    def __repr__(self) -> str:
        return f"Feature({self._name!r}, {self._type!r}, {self._description!r}, {self._primary}, {self._foreign}, {self._partition}, {self._online_type!r}, {self._default_value!r}, {self._feature_group_id!r})"

    def __hash__(self) -> int:
        return hash(f"{self.feature_group_id}_{self.name}")



#
#   Copyright 2021 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from hopsworks_common.util import (
    Encoder,
    NumpyEncoder,
    ProvenanceWarning,
    VersionWarning,
    _handle_dataframe_input,
    _handle_dict_input,
    _handle_tensor_input,
    compress,
    decompress,
    extract_field_from_json,
    feature_view_to_json,
    get_hostname_replaced_url,
    get_members,
    get_obj_from_json,
    get_predictor_for_model,
    input_example_to_json,
    pretty_print,
    set_model_class,
    validate_metrics,
)


MLEncoder = Encoder


__all__ = [
    "Encoder",
    "MLEncoder",
    "NumpyEncoder",
    "ProvenanceWarning",
    "VersionWarning",
    "_handle_dataframe_input",
    "_handle_dict_input",
    "_handle_tensor_input",
    "compress",
    "decompress",
    "extract_field_from_json",
    "feature_view_to_json",
    "get_hostname_replaced_url",
    "get_members",
    "get_obj_from_json",
    "get_predictor_for_model",
    "input_example_to_json",
    "pretty_print",
    "set_model_class",
    "validate_metrics",
]



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

from typing import List, Optional, Union

from hopsworks_common import client
from hsfs import decorators
from hsfs.core import feature_monitoring_config as fmc
from hsfs.core.job import Job


class FeatureMonitoringConfigApi:
    """Feature Monitoring Configuration endpoints for the Feature Group resource.
    :param feature_store_id: id of the respective Feature Store
    :type feature_store_id: int
    :param feature_group_id: id of the feature group, if monitoring a feature group
    :type feature_group_id: int, optional
    :param feature_view_name: name of the feature view, if monitoring a feature view
    :type feature_view_name: str, optional
    :param feature_view_version: version of the feature view, if monitoring a feature view
    :type feature_view_version: int, optional
    """

    def __init__(
        self,
        feature_store_id: int,
        feature_group_id: Optional[int] = None,
        feature_view_name: Optional[str] = None,
        feature_view_version: Optional[int] = None,
        **kwargs,
    ) -> None:
        if feature_group_id is None:
            assert feature_view_name is not None
            assert feature_view_version is not None

        self._feature_store_id = feature_store_id
        self._feature_group_id = feature_group_id
        self._feature_view_name = feature_view_name
        self._feature_view_version = feature_view_version

    def create(
        self,
        fm_config: fmc.FeatureMonitoringConfig,
    ) -> fmc.FeatureMonitoringConfig:
        """Create an feature monitoring configuration attached to the Feature of a Feature Group.
        :param fm_config: feature monitoring config object to be attached to a Feature
        :type fm_config: `FeatureMonitoringConfiguration`
        :return: the created feature monitoring configuration
        :rtype: FeatureMonitoringConfiguration
        """
        _client = client.get_instance()
        path_params = self.build_path_params(
            project_id=_client._project_id,
        )

        headers = {"content-type": "application/json"}
        payload = fm_config.json()
        return fmc.FeatureMonitoringConfig.from_response_json(
            _client._send_request("POST", path_params, headers=headers, data=payload)
        )

    def update(
        self,
        fm_config: fmc.FeatureMonitoringConfig,
    ) -> fmc.FeatureMonitoringConfig:
        """Update a feature monitoring configuration attached to a Feature.
        :param fm_config: feature monitoring configuration to be attached to a Feature
        :type fm_config: `FeatureMonitoringConfig`
        :return: the updated feature monitoring configuration
        :rtype: FeatureMonitoringConfig
        """
        _client = client.get_instance()
        path_params = self.build_path_params(
            project_id=_client._project_id,
            config_id=fm_config._id,
        )

        headers = {"content-type": "application/json"}
        payload = fm_config.json()

        return fmc.FeatureMonitoringConfig.from_response_json(
            _client._send_request("PUT", path_params, headers=headers, data=payload)
        )

    def delete(
        self,
        config_id: int,
    ) -> None:
        """Delete the Feature Monitoring configuration attached to a Feature.
        :param config_id: Id of the feature monitoring configuration to delete
        :type config_id: int
        :return: fetched feature monitoring configuration attached to the Feature Group
        :rtype: FeatureMonitoringConfig || None
        """
        _client = client.get_instance()
        path_params = self.build_path_params(
            project_id=_client._project_id,
            config_id=config_id,
        )

        _client._send_request("DELETE", path_params)

    def get_by_id(
        self,
        config_id: int,
    ) -> Optional[fmc.FeatureMonitoringConfig]:
        """Get the Feature Monitoring Configuration attached to a Feature.
        :param config_id: Id of the feature monitoring configuration to fetch
        :type config_id: int
        :return: fetched feature monitoring configuration attached to the Feature Group
        :rtype: FeatureMonitoringConfig || None
        """
        _client = client.get_instance()
        path_params = self.build_path_params(
            project_id=_client._project_id,
            config_id=config_id,
        )

        return fmc.FeatureMonitoringConfig.from_response_json(
            _client._send_request("GET", path_params)
        )

    def get_by_feature_name(
        self,
        feature_name: str,
    ) -> Optional[fmc.FeatureMonitoringConfig]:
        """Get all Feature Monitoring Configurations attached to a Feature Name.
        :param feature_name: Name of the feature for which to fetch monitoring configuration
        :type feature_name: str
        :return: fetched feature monitoring configuration attached to the Feature Group
        :rtype: List[FeatureMonitoringConfig] || None
        """
        _client = client.get_instance()
        path_params = self.build_path_params(
            project_id=_client._project_id,
            feature_name=feature_name,
        )

        return fmc.FeatureMonitoringConfig.from_response_json(
            _client._send_request("GET", path_params)
        )

    @decorators.catch_not_found(
        "hsfs.core.feature_monitoring_config.FeatureMonitoringConfig",
        fallback_return=[],
    )
    def get_by_name(
        self,
        name: str,
    ) -> Optional[fmc.FeatureMonitoringConfig]:
        """Get all Feature Monitoring Configurations attached to a Feature Name.
        :param name: Name of the feature monitoring configuration to fetch
        :type name: str
        :return: fetched feature monitoring configuration attached to the Feature Group
        :rtype: FeatureMonitoringConfig || None
        """
        _client = client.get_instance()
        path_params = self.build_path_params(
            project_id=_client._project_id,
            name=name,
        )
        return fmc.FeatureMonitoringConfig.from_response_json(
            _client._send_request("GET", path_params)
        )

    def get_by_entity(self) -> List[fmc.FeatureMonitoringConfig]:
        """Get all Feature Monitoring Configurations attached to a Feature Name.
        :param name: Name of the feature monitoring configuration to fetch
        :type name: str
        :return: fetched feature monitoring configuration attached to the Feature Group
        :rtype: List[FeatureMonitoringConfig] || None
        """
        _client = client.get_instance()
        path_params = self.build_path_params(
            project_id=_client._project_id,
            entity=True,
        )

        return fmc.FeatureMonitoringConfig.from_response_json(
            _client._send_request("GET", path_params)
        )

    def setup_feature_monitoring_job(
        self,
        config_name: str,
    ) -> Job:
        """Setup a feature monitoring job for a configuration.
        :param config_name: Name of the feature monitoring configuration to setup a job for
        :type config_name: str
        :return: Job object for the feature monitoring job
        :rtype: Job
        """
        _client = client.get_instance()
        path_params = self.build_path_params(
            project_id=_client._project_id,
        )
        path_params.extend(["setup", config_name])

        return Job.from_response_json(_client._send_request("POST", path_params))

    def trigger_feature_monitoring_job(
        self,
        config_id: int,
    ) -> Job:
        """Trigger a feature monitoring job for a configuration.
        :param config_id: Id of the feature monitoring configuration to trigger a job for
        :type config_id: int
        :return: Job attached to the monitoring configuration
        :rtype: Job
        """
        _client = client.get_instance()
        path_params = self.build_path_params(
            project_id=_client._project_id,
            config_id=config_id,
        )
        path_params.append("trigger")

        return Job.from_response_json(_client._send_request("POST", path_params))

    def build_path_params(
        self,
        project_id: int,
        feature_name: Optional[str] = None,
        config_id: Optional[int] = None,
        name: Optional[str] = None,
        entity: bool = False,
    ) -> List[Union[str, int]]:
        """Builds the path parameters for the Feature Monitoring Config API.
        :param project_id: Id of the project
        :type project_id: int
        :type feature_name: str, optional
        :param config_id: Id of the feature monitoring configuration.
            Only to fetch feature monitoring configuration by id.
        :type config_id: int, optional
        :param name: Name of the feature monitoring configuration.
            Only to fetch feature monitoring configuration by name.
        :type name: str, optional
        :param entity: Whether to append entity to path params. Defaults to False.
            Only to fetch all feature monitoring configurations attached to an entity.
        :return: path parameters
        :rtype: List[str]
        """
        path_params = [
            "project",
            project_id,
            "featurestores",
            self._feature_store_id,
        ]
        if self._feature_group_id is not None:
            path_params.extend(["featuregroups", self._feature_group_id])
        else:
            path_params.extend(
                [
                    "featureview",
                    self._feature_view_name,
                    "version",
                    self._feature_view_version,
                ]
            )
        path_params.extend(["featuremonitoring", "config"])

        if config_id:
            path_params.append(config_id)
        elif feature_name:
            path_params.extend(["feature", feature_name])
        elif name:
            path_params.extend(["name", name])
        elif entity:
            path_params.append("entity")

        return path_params



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

from typing import List

from hopsworks_common import client
from hsfs import feature_store_activity as fsa_mod


class FeatureStoreActivityApi:
    def get_feature_group_activities(
        self,
        feature_store_id: int,
        feature_group_id: int,
        activity_type: List[fsa_mod.FeatureStoreActivityType],
        limit: int = 100,
        offset: int = 0,
    ) -> List[fsa_mod.FeatureStoreActivity]:
        """
        Get the activities for a feature group in the feature store
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            feature_store_id,
            "featuregroups",
            feature_group_id,
            "activity",
        ]

        query_params = {
            "limit": limit,
            "offset": offset,
            "expand": [
                "users",
                "commits",
                "jobs",
                "validationreport",
                "expectationsuite",
                "executions",
                "statistics",
            ],
            "sortby": "timestamp:desc",
        }
        if activity_type is not None and len(activity_type) > 0:
            query_params["activityType"] = activity_type

        response = _client._send_request("GET", path_params, query_params=query_params)

        return fsa_mod.FeatureStoreActivity.from_response_json(response)



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from hopsworks_common.client.exceptions import (
    DatasetException,
    DataValidationException,
    EnvironmentException,
    ExternalClientError,
    FeatureStoreException,
    GitException,
    InternalClientError,
    JobException,
    JobExecutionException,
    KafkaException,
    ModelRegistryException,
    ModelServingException,
    OpenSearchException,
    ProjectException,
    RestAPIError,
    UnknownSecretStorageError,
    VectorDatabaseException,
)


__all__ = [
    "DatasetException",
    "DataValidationException",
    "EnvironmentException",
    "ExternalClientError",
    "FeatureStoreException",
    "GitException",
    "InternalClientError",
    "JobException",
    "JobExecutionException",
    "KafkaException",
    "ModelRegistryException",
    "ModelServingException",
    "OpenSearchException",
    "ProjectException",
    "RestAPIError",
    "UnknownSecretStorageError",
    "VectorDatabaseException",
]



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import copy

import humps
from hsml import inference_batcher
from hsml.constants import INFERENCE_BATCHER


class TestInferenceBatcher:
    # from response json

    def test_from_response_json_enabled(self, mocker, backend_fixtures):
        # Arrange
        json = backend_fixtures["inference_batcher"]["get_enabled"]["response"]
        json_camelized = humps.camelize(json)  # as returned by the backend
        mock_ib_from_json = mocker.patch(
            "hsml.inference_batcher.InferenceBatcher.from_json"
        )

        # Act
        _ = inference_batcher.InferenceBatcher.from_response_json(json_camelized)

        # Assert
        mock_ib_from_json.assert_called_once_with(json)

    def test_from_response_json_enabled_with_config(self, mocker, backend_fixtures):
        # Arrange
        json = backend_fixtures["inference_batcher"]["get_enabled_with_config"][
            "response"
        ]
        json_camelized = humps.camelize(json)  # as returned by the backend
        mock_ib_from_json = mocker.patch(
            "hsml.inference_batcher.InferenceBatcher.from_json"
        )

        # Act
        _ = inference_batcher.InferenceBatcher.from_response_json(json_camelized)

        # Assert
        mock_ib_from_json.assert_called_once_with(json)

    # from json

    def test_from_json_enabled(self, mocker, backend_fixtures):
        # Arrange
        json = backend_fixtures["inference_batcher"]["get_enabled"]["response"]
        mock_ib_extract_fields = mocker.patch(
            "hsml.inference_batcher.InferenceBatcher.extract_fields_from_json",
            return_value=json,
        )
        mock_ib_init = mocker.patch(
            "hsml.inference_batcher.InferenceBatcher.__init__", return_value=None
        )

        # Act
        _ = inference_batcher.InferenceBatcher.from_json(json)

        # Assert
        mock_ib_extract_fields.assert_called_once_with(json)
        mock_ib_init.assert_called_once_with(**json)

    def test_from_json_enabled_with_config(self, mocker, backend_fixtures):
        # Arrange
        json = backend_fixtures["inference_batcher"]["get_enabled_with_config"][
            "response"
        ]
        mock_ib_extract_fields = mocker.patch(
            "hsml.inference_batcher.InferenceBatcher.extract_fields_from_json",
            return_value=json,
        )
        mock_ib_init = mocker.patch(
            "hsml.inference_batcher.InferenceBatcher.__init__", return_value=None
        )

        # Act
        _ = inference_batcher.InferenceBatcher.from_json(json)

        # Assert
        mock_ib_extract_fields.assert_called_once_with(json)
        mock_ib_init.assert_called_once_with(**json)

    # constructor

    def test_constructor_default(self):
        # Act
        ib = inference_batcher.InferenceBatcher()

        # Assert
        assert isinstance(ib, inference_batcher.InferenceBatcher)
        assert ib.enabled == INFERENCE_BATCHER.ENABLED
        assert ib.max_batch_size is None
        assert ib.max_latency is None
        assert ib.timeout is None

    def test_constructor_enabled(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["inference_batcher"]["get_enabled"]["response"]

        # Act
        ib = inference_batcher.InferenceBatcher(**json)

        # Assert
        assert isinstance(ib, inference_batcher.InferenceBatcher)
        assert ib.enabled == json["enabled"]
        assert ib.max_batch_size is None
        assert ib.max_latency is None
        assert ib.timeout is None

    def test_constructor_enabled_with_config(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["inference_batcher"]["get_enabled_with_config"][
            "response"
        ]

        # Act
        ib = inference_batcher.InferenceBatcher(**json)

        # Assert
        assert isinstance(ib, inference_batcher.InferenceBatcher)
        assert ib.enabled == json["enabled"]
        assert ib.max_batch_size == json["max_batch_size"]
        assert ib.max_latency == json["max_latency"]
        assert ib.timeout == json["timeout"]

    def test_constructor_disabled(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["inference_batcher"]["get_disabled"]["response"]

        # Act
        ib = inference_batcher.InferenceBatcher(**json)

        # Assert
        assert isinstance(ib, inference_batcher.InferenceBatcher)
        assert ib.enabled == json["enabled"]
        assert ib.max_batch_size is None
        assert ib.max_latency is None
        assert ib.timeout is None

    def test_constructor_disabled_with_config(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["inference_batcher"]["get_disabled_with_config"][
            "response"
        ]

        # Act
        ib = inference_batcher.InferenceBatcher(**json)

        # Assert
        assert isinstance(ib, inference_batcher.InferenceBatcher)
        assert ib.enabled == json["enabled"]
        assert ib.max_batch_size == json["max_batch_size"]
        assert ib.max_latency == json["max_latency"]
        assert ib.timeout == json["timeout"]

    # # extract fields from json

    def test_extract_fields_from_json_enabled(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["inference_batcher"]["get_enabled"]["response"]
        json_copy = copy.deepcopy(json)

        # Act
        kwargs = inference_batcher.InferenceBatcher.extract_fields_from_json(json_copy)

        # Assert
        assert kwargs["enabled"] == json["enabled"]
        assert kwargs["max_batch_size"] is None
        assert kwargs["max_latency"] is None
        assert kwargs["timeout"] is None

    def test_extract_fields_from_json_enabled_with_config(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["inference_batcher"]["get_enabled_with_config"][
            "response"
        ]
        json_copy = copy.deepcopy(json)

        # Act
        kwargs = inference_batcher.InferenceBatcher.extract_fields_from_json(json_copy)

        # Assert
        assert kwargs["enabled"] == json["enabled"]
        assert kwargs["max_batch_size"] == json["max_batch_size"]
        assert kwargs["max_latency"] == json["max_latency"]
        assert kwargs["timeout"] == json["timeout"]

    def test_extract_fields_from_json_enabled_nested(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["inference_batcher"]["get_enabled"]["response_nested"]
        json_copy = copy.deepcopy(json)

        # Act
        kwargs = inference_batcher.InferenceBatcher.extract_fields_from_json(json_copy)

        # Assert
        assert kwargs["enabled"] == json["batching_configuration"]["enabled"]
        assert kwargs["max_batch_size"] is None
        assert kwargs["max_latency"] is None
        assert kwargs["timeout"] is None

    def test_extract_fields_from_json_enabled_with_config_nested(
        self, backend_fixtures
    ):
        # Arrange
        json = backend_fixtures["inference_batcher"]["get_enabled_with_config"][
            "response_nested"
        ]
        json_copy = copy.deepcopy(json)

        # Act
        kwargs = inference_batcher.InferenceBatcher.extract_fields_from_json(json_copy)

        # Assert
        assert kwargs["enabled"] == json["batching_configuration"]["enabled"]
        assert (
            kwargs["max_batch_size"] == json["batching_configuration"]["max_batch_size"]
        )
        assert kwargs["max_latency"] == json["batching_configuration"]["max_latency"]
        assert kwargs["timeout"] == json["batching_configuration"]["timeout"]



#
#   Copyright 2025 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import json
import time
import warnings
from datetime import datetime, timedelta
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Union,
)

import humps
from hopsworks_common import client, util
from hsfs import feature_group as fg_mod
from hsfs.core import online_ingestion_result
from hsfs.core.opensearch import OpenSearchClientSingleton
from tqdm.auto import tqdm


class OnlineIngestion:
    """
    Metadata object used to provide Online Ingestion information for a feature group.
    """

    def __init__(
        self,
        id: Optional[int] = None,
        num_entries: Optional[int] = None,
        results: Union[
            List[online_ingestion_result.OnlineIngestionResult],
            List[Dict[str, Any]],
        ] = None,
        feature_group: fg_mod.FeatureGroup = None,
        **kwargs,
    ):
        self._id = id
        self._num_entries = num_entries  # specified when inserting (optional since might not be specified when using streaming)
        self._results = (
            [
                (
                    online_ingestion_result.OnlineIngestionResult.from_response_json(
                        result
                    )
                    if isinstance(result, dict)
                    else result
                )
                for result in results
            ]
            if results
            else []
        )  # batch inserts performed by onlinefs
        self._feature_group = feature_group

    @classmethod
    def from_response_json(
        cls, json_dict: Dict[str, Any], feature_group: fg_mod.FeatureGroup = None
    ) -> OnlineIngestion:
        if json_dict is None:
            return None

        json_decamelized: dict = humps.decamelize(json_dict)

        if "count" not in json_decamelized:
            return cls(**json_decamelized, feature_group=feature_group)
        elif json_decamelized["count"] == 1:
            return cls(**json_decamelized["items"][0], feature_group=feature_group)
        elif json_decamelized["count"] > 1:
            return [
                cls(**item, feature_group=feature_group)
                for item in json_decamelized["items"]
            ]
        else:
            return None

    def refresh(self):
        online_ingestion = self.feature_group.get_online_ingestion(self._id)
        self.__dict__.update(online_ingestion.__dict__)

    def to_dict(self):
        return {"id": self._id, "numEntries": self._num_entries}

    def json(self):
        return json.dumps(self, cls=util.Encoder)

    @property
    def id(self) -> Optional[int]:
        return self._id

    @property
    def num_entries(self) -> Optional[int]:
        return self._num_entries

    @num_entries.setter
    def num_entries(self, num_entries: int) -> None:
        self._num_entries = num_entries

    @property
    def results(
        self,
    ) -> List[online_ingestion_result.OnlineIngestionResult]:
        return self._results

    @property
    def feature_group(self) -> fg_mod.FeatureGroup:
        return self._feature_group

    def wait_for_completion(self, options: Dict[str, Any] = None):
        if options is None:
            options = {}

        # Set timeout time
        timeout_delta = timedelta(seconds=options.get("timeout", 60))
        timeout_time = datetime.now() + timeout_delta

        with tqdm(
            total=self.num_entries,
            bar_format="{desc}: {percentage:.2f}% |{bar}| Rows {n_fmt}/{total_fmt}",
            desc="Online data ingestion progress",
            mininterval=1,
        ) as progress_bar:
            while True:
                # Get total number of rows processed
                rows_processed = sum(result.rows for result in self.results)

                # Update progress bar
                if any(result.status != "UPSERTED" for result in self.results):
                    progress_bar.colour = "RED"
                progress_bar.n = rows_processed
                progress_bar.refresh()

                # Check if the online ingestion is complete
                if self.num_entries and rows_processed >= self.num_entries:
                    break

                # Check if the timeout has been reached (if timeout is 0 we will wait indefinitely)
                if timeout_delta != timedelta(0) and datetime.now() >= timeout_time:
                    warnings.warn(
                        f"Timeout of {timeout_delta} was exceeded while waiting for online ingestion completion.",
                        stacklevel=1,
                    )
                    break

                # Sleep for the specified period in seconds
                time.sleep(options.get("period", 1))

                self.refresh()

    def print_logs(self, priority: str = "error", size: int = 20):
        open_search_client = OpenSearchClientSingleton()

        response = open_search_client.search(
            body={
                "query": {
                    "bool": {
                        "must": [
                            {
                                "match": {
                                    "log_arguments.feature_group_id": f"{self.feature_group.id}"
                                }
                            },
                            {
                                "match": {
                                    "log_arguments.online_ingestion_id": f"{self.id}"
                                }
                            },
                            {"match": {"priority": priority}},
                        ]
                    }
                },
                "size": size,
            },
            index=f"onlinefs_{client.get_instance()._project_id}-*",
        )

        for hit in response["hits"]["hits"]:
            print(hit["_source"]["error"]["data"])
